/*
 * secfingerd
 * secfingerd.c - secure, user customizable finger daemon
 *
 * Copyright (c) 1999 Ni9 Technologies [x3/intrinsic]
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "./secfingerd.h"
#include "./config.h"

int
main(argc, argv)
	int argc;
	char **argv;
{
	int arg, log_level, template_flag, lookup_flag, len;
	unsigned long remote;
	char *global_nofinger_file_name, query[LENGTH_QUERY], *query_p,
	  *full_path, *default_template_file_name, *remote_hostname,
	  remote_numerical[LENGTH_IP];
	struct sockaddr_in sock_info;
	struct passwd *user_info;
	struct hostent *host_info;
	
	/*
	 * Set the flags and variables that may be changed by command line
	 * arguments to their default values before reading the arguments and
	 * changing them depending on the arguments.
	 */
	log_level = 0;
	template_flag = 0;
	lookup_flag = 0;
	default_template_file_name = NULL;
	global_nofinger_file_name = G_NOFINGER;
	
	while((arg = getopt(argc, argv, "hvkl:g:t:")) != EOF) {
		switch(arg) {
			case 'h':
				usage(argv[0]);
				exit(0);
			case 'v':
				fprintf(stderr, "%s\n", VERSION);
				exit(0);
			case 'k':
				lookup_flag = 1;
				break;
			case 'l':
				log_level = atoi(optarg);
				if((log_level != 1) && (log_level != 2))
					exit(1);
				break;
			case 'g':
				global_nofinger_file_name = optarg;
				break;
			case 't':
				template_flag = 1;
				default_template_file_name = optarg;
				break;
			default:
				exit(1);
		}
	}

	/*
	 * If logging is on then open the syslog connection.
	 */
	if(log_level > 0)
		openlog("secfingerd", LOG_PID, LOG_DAEMON);
	
	/*
	 * Get the client's IP address, if it is not a valid socket then exit
	 * with an error.  Also get the query.
	 */
	len = sizeof(sock_info);
	if(getpeername(0, (struct sockaddr *) &sock_info, &len) < 0) {
		if(errno = ENOTSOCK)
			remote = 0x7f000001;
		else
			exit(1);
	} else {
		if(sock_info.sin_family != AF_INET)
			exit(1);
		else
			remote = ntohl(sock_info.sin_addr.s_addr);
	}
	
	if(!fgets(query, LENGTH_QUERY, stdin))
		exit(1);

	/*
	 * Set the remote_numerical string to the numerical IP address of the
	 * remote client.  If the -k flag was used then set the remote_hostname
	 * string to the client's hostname, on failure set it to K_FAILURE.  If
	 * -k was not used then set remote_hostname to K_NOTCHOSEN.
	 */
	snprintf(remote_numerical, LENGTH_IP, "%ld.%ld.%ld.%ld", remote>>24 &
	  255, remote>>16 &255, remote>>8 & 255, remote>>16 & 255);
	if(lookup_flag == 1) {
		if((host_info = gethostbyaddr((char *) &sock_info.sin_addr,
		  sizeof(struct in_addr), AF_INET)))
		  	remote_hostname = host_info->h_name;
		else
			remote_hostname = K_FAILURE;
	} else {
		remote_hostname = K_NOTCHOSEN;
	}

	/*
	 * Clean up the query and check for empty queries and indirect queries,
	 * if an empty query is recieved then send the E_EMPTYQ message to the
	 * client and if an indirect query is recieved then send an E_INDIRECTQ
	 * message to the client.  If either of these are recieved and logging
	 * is on these will be logged.
	 */
	for(len = 0; query[len]; len++) {
		if((query[len] == '\r') || (query[len] == '\n')) {
			query[len] = '\0';
			break;
		}
		if(query[len] == '@') {
			printf("%s\n", E_INDIRECTQ);
			if(log_level > 0) {
				syslog(LOG_WARNING, "Indirect query from "     \
			  	  "\"%s\" (\"%s\").", remote_numerical,
			  	  remote_hostname);
				closelog();
			}
			exit(1);
		}
	}
	
	query_p = query;
	
	if(*query_p == ' ')
		query_p++;
	if((*query_p == '/') && ((*(query_p + 1) == 'W') || (*(query_p + 1) ==
	  'w')))
	  	query_p += 2;
	if(*query_p == ' ')
		query_p++;
	if(*query_p == 0) {
		printf("%s\n", E_EMPTYQ);
		if(log_level > 0) {
			syslog(LOG_WARNING, "Empty query from "                \
			  "\"%s\" (\"%s\").", remote_numerical,
			  remote_hostname);
			closelog();
		}
		exit(1);
	}
	
	/*
	 * If logging is on then log the query along with the numerical IP
	 * address of the client.
	 */
	if(log_level > 1)
		syslog(LOG_INFO, "Query for information on \"%s\" from "       \
		  "\"%s\" (\"%s\").", query_p, remote_numerical,
		  remote_hostname);
	
	/*
	 * Check that the user exists by looking up the user name in the
	 * password file, if the user does not exist then send the E_NOUSER
	 * error message to the client.
	 */
	if((user_info = getpwnam(query_p))) {
		/*
		 * Check if the user has a nofinger file in their home directory
		 * or their user name is in the global nofinger file, if they do
		 * not want to be fingered send the E_NOFINGER message to the
		 * client.
		 */
		if((check_for_nofinger(user_info, global_nofinger_file_name))
		  == 0) {
			/*
			 * Check if the user has a template file in their home
			 * directory, if they do not then use the default
			 * template.
			 */
			if((check_for_template(user_info)) == 0) {
				/*
				 * If NO_SYMLINKS options is set to 1 then check
				 * that the template file is not a symbolic
				 * link.
				 */
				full_path = build_full_path(user_info,
				  U_TEMPLATE);
				if((NO_SYMLINKS == 1) && (check_for_symlink(
				  full_path) == -1)) {
					printf("%s\n", E_SYMLINK);
					if(log_level > 0) {
						syslog(LOG_WARNING, "Symbolic" \
						  " link found in \"%s\", "    \
						  "query from \"%s\" (\"%s\").",
						  full_path, remote_numerical,
						  remote_hostname);
					}
					closelog();
					exit(1);
				}
				
				dump_using_template(user_info, full_path,
				  log_level, remote_numerical,
				  remote_hostname);
			} else {
				/*
				 * If no custom template for users who do not
				 * have their own custom templates has been
				 * specified then use the internal format,
				 * otherwise use the custom one.
				 */
				if(template_flag == 1) {
					dump_using_template(user_info,
					  default_template_file_name, log_level,
					  remote_numerical, remote_hostname);
				} else {
					dump_without_template(user_info,
					  log_level, remote_numerical,
					  remote_hostname);
				}
			}
		} else {
			printf("%s\n", E_NOFINGER);
			closelog();
			exit(1);
		}
	} else {
		printf("%s\n", E_NOUSER);
		closelog();
		exit(1);
	}
	
	/*
	 * If logging is on then close the syslog connection.
	 */
	if(log_level > 0)
		closelog();
}

void
usage(name)
	char *name;
{
	fprintf(stderr, "Usage: %s [-hv] [-k] [-l level] [-g file] [-t file]\n",
	  name);
	fprintf(stderr, "  -h        display the correct program usage\n");
	fprintf(stderr, "  -v        display the program version\n");
	fprintf(stderr, "  -k        log the hostname of the client as well"   \
	  " as the numerical IP address\n");
	fprintf(stderr, "  -l level  turn logging on and set the level.  In"   \
	  " level 1 indirect and\n            empty queries are logged.  In"   \
	  " level 2 every query is\n            logged (including indirect"    \
	  " and empty queries)\n");
	fprintf(stderr, "  -g file   change the global nofinger file from the" \
	  " default one which was\n            set in \"config.h\"\n");
	fprintf(stderr, "  -t file   set a template file for users who do not" \
	  " have a custom template\n            file in their home directory," \
	  " if this option is not\n            used and the user does not have"\
	  " a custom template file\n            then the hard coded one in the"\
	  " source code is used\n");
}

char *
build_full_path(user_info, filename)
	char *filename;
	struct passwd *user_info;
{
	static char short_filename[LENGTH_PATH], short_directory[LENGTH_PATH],
	  full_path[(2 * LENGTH_PATH) + 30];
	
	strncpy(short_filename, filename, LENGTH_PATH);
	strncpy(short_directory, user_info->pw_dir, LENGTH_PATH);
	snprintf(full_path, sizeof(full_path), "%s/%s", short_directory,
	  short_filename);
	
	return(full_path);
}

int
read_user(file_handle, user)
	char *user;
	FILE *file_handle;
{
	int i, byte;
	char tmp_user[LENGTH_USER];
	
	i = 0;
	while(i < LENGTH_USER) {
		byte = fgetc(file_handle);
		if((byte == ' ') || (byte == '\n') || (byte == '\r') ||
		  (byte == EOF))
			break;
		tmp_user[i] = byte;
		i++;
	}
	tmp_user[i] = 0;
	
	if(byte == EOF)
		return(EOF);
	
	strncpy(user, tmp_user, LENGTH_USER);
	return(0);
}

int
check_for_nofinger(user_info, global_nofinger_file_name)
	char *global_nofinger_file_name;
	struct passwd *user_info;
{
	char *full_path, user[LENGTH_USER];
	FILE *file_handle;
	
	full_path = build_full_path(user_info, U_NOFINGER);
	if((file_handle = fopen(full_path, "r")) == NULL) {
		if((file_handle = fopen(global_nofinger_file_name, "r")) !=
		  NULL) {
			while((read_user(file_handle, (char *) &user)) != EOF) {
				if(strcmp(user, user_info->pw_name) == 0) {
					fclose(file_handle);
					return(-1);
				}
			}
			fclose(file_handle);
		} 
	} else {
		fclose(file_handle);
		return(-1);
	}
	
	return(0);
}

int
check_for_template(user_info)
	struct passwd *user_info;
{
	char *full_path;
	FILE *file_handle;
	
	full_path = full_path = build_full_path(user_info, U_TEMPLATE);
	if((file_handle = fopen(full_path, "r")) == NULL) {
		return(-1);
	} else {
		fclose(file_handle);
		return(0);
	}
}

int
check_for_symlink(filename)
	char *filename;
{
	struct stat file_info;
	
	lstat(filename, &file_info);
	if(S_ISLNK(file_info.st_mode))
		return(-1);
	else
		return(0);
}

void
dump_file(user_info, filename, log_level, remote_numerical, remote_hostname)
	int log_level;
	char *filename, *remote_numerical, *remote_hostname;
	struct passwd *user_info;
{
	int byte;
	char *full_path;
	FILE *file_handle;
	
	full_path = build_full_path(user_info, filename);
	if((file_handle = fopen(full_path, "r")) == NULL) {
		printf("%s\n", E_FILENF);
	} else {
		if((NO_SYMLINKS == 1) && (check_for_symlink(full_path) == -1)) {
			printf("%s\n", E_SYMLINK);
			if(log_level > 0)
				syslog(LOG_WARNING, "Symbolic link found in "  \
				  "\"%s\", query from \"%s\" (\"%s\").",
				  full_path, remote_numerical, remote_hostname);
		} else {
			while(byte = fgetc(file_handle), byte != EOF)
				putchar(byte);
		}
		fclose(file_handle);
	}
}

void
dump_using_template(user_info, filename, log_level, remote_numerical,
  remote_hostname)
  	int log_level;
	char *filename, *remote_numerical, *remote_hostname;
	struct passwd *user_info;
{
	int byte1, byte2;
	FILE *file_handle;
	
	if((file_handle = fopen(filename, "r")) == NULL) {
		printf("The template file existed a minute ago... honest\n");
		closelog();
		exit(1);
	}
	
	while(byte1 = fgetc(file_handle), byte1 != EOF) {
		if(byte1 == '\\') {
			if((byte2 = fgetc(file_handle)) == EOF)
				return;
			if(byte2 == '\\')
				putchar('\\');
			else if(byte2 == S_NAME)
				printf("%s", user_info->pw_name);
			else if(byte2 == S_GECOS)
				printf("%s", user_info->pw_gecos);
			else if(byte2 == S_DIR)
				printf("%s", user_info->pw_dir);
			else if(byte2 == S_SHELL)
				printf("%s", user_info->pw_shell);
			else if(byte2 == S_PLAN)
				dump_file(user_info, U_PLAN, log_level,
				  remote_numerical, remote_hostname);
			else if(byte2 == S_PROJECT)
				dump_file(user_info, U_PROJECT, log_level,
				  remote_numerical, remote_hostname);
			else if(byte2 == S_PUBKEY)
				dump_file(user_info, U_PUBKEY, log_level,
				  remote_numerical, remote_hostname);
			else if(byte2 == 'w')
				printf("Carefull you do not get ambushed by"   \
				  " giant wood elves.");
			else if(byte2 == 'h')
				printf("Come, come quick you can see our"      \
				  " house from here.");
			else if(byte2 == 'l')
				printf("Life will never be the same in Port"   \
				  " Royal town.");
		} else {
			putchar(byte1);
		}
	}
	
	fclose(file_handle);
}

void
dump_without_template(user_info, log_level, remote_numerical, remote_hostname)
	int log_level;
	char *remote_numerical, *remote_hostname;
	struct passwd *user_info;
{
	printf("Username: %s\n", user_info->pw_name);
	printf("Realname: %s\n", user_info->pw_gecos);
	printf("-------------------------------------------------------------" \
	  "-------------------\n");
	printf("Plan:\n");
	dump_file(user_info, U_PLAN, log_level, remote_numerical,
	  remote_hostname);
	printf("\nProject:\n");
	dump_file(user_info, U_PROJECT, log_level, remote_numerical,
	  remote_hostname);
	printf("\nPublic Key:\n");
	dump_file(user_info, U_PUBKEY, log_level, remote_numerical,
	  remote_hostname);
}
